// vMac 0.1 Source
// By Philip Cummins (PC)
// Copyright PC 1997

// PC - SCC Function Files

// Zilog Z8530 (SCC) emulation
// Started by PC, minimal emulation
// and serial support by Weston

#include "sysconfi.h"
#include "sysdeps.h"
#include "memory2.h"
#include "config.h"
#include "SCC.h"
#include "Prefs.h"

#ifdef _SCC_Debug
#include <stdio.h>
#endif

#include "IBM-SCC.h"

FILE *SCC_A_Out;
FILE *SCC_B_Out;

//Just to make things a little easier
#define Bit0 1
#define Bit1 2
#define Bit2 4
#define Bit3 8
#define Bit4 16
#define Bit5 32
#define Bit6 64
#define Bit7 128

//SCC Interrupts
#define SCC_A_Rx       8 //Rx Char Available
#define SCC_A_Rx_Spec  7 //Rx Special Condition
#define SCC_A_Tx_Empty 6 //Tx Buffer Empty
#define SCC_A_Ext      5 //External/Status Change
#define SCC_B_Rx       4 //Rx Char Available
#define SCC_B_Rx_Spec  3 //Rx Special Condition
#define SCC_B_Tx_Empty 2 //Tx Buffer Empty
#define SCC_B_Ext      1 //External/Status Change


SCC_Ty SCC;
UBYTE SCC_Mode = 0;
UBYTE SCC_Reg = 0;

int  SCC_Interrupt_Type;

void SCC_Interrupt (int Type);
void SCC_PutReg    (int in, int chan);
void SCC_ResetB    (void);
void SCC_ResetA    (void);

PAR_HANDLE PrintPort;
char PrintData[16384];
int PrintBytes,PrintCount,ReadPrint=0;

SER_HANDLE ModemPort;
char ModemData[16384];
int ModemBytes,ModemCount,ReadModem=0;

int SCC_Init (void)
{
   SCC_Mode = SCC_Reg = SCC_Interrupt_Type = 0;

   SCC_ResetB();
   SCC_ResetA();

   if (thePrefs.Printer) {
      PrintPort = Printer_InitPort(thePrefs.Printer);
      if (PrintPort.PortNumber == 0xff) {
         fprintf(stderr, "vMac: Cannot open printer port.\n");
         fprintf(stderr, "vMac: Sorry, errors encountered - cannot continue.\n");
         exit(1);
      }
   } else {
      PrintPort.PortNumber = 0xff;
   }

   ModemPort.PortNumber = 0xff;

   return 0;
}

void  SCC_Reset (void)
{
   (void) SCC_Init();
   return;
}

void SCC_ResetB(void)
{
   int Counter;

   for (Counter = 0; Counter < 16; ++Counter) {
       SCC.B.RR[Counter] = 0;
       SCC.B.WR[Counter] = 0;
   }

   SCC.B.RR[0] = Bit5+Bit4+Bit3+Bit2;
   SCC.B.RR[1] = Bit2+Bit1+Bit0;
   ReadPrint=0;

   return;
}

void SCC_ResetA(void)
{
   int Counter;

   for (Counter = 0; Counter < 16; ++Counter) {
      SCC.A.RR[Counter] = 0;
      SCC.A.WR[Counter] = 0;
   }

   SCC.A.RR[0] = Bit5+Bit4+Bit3+Bit2;
   SCC.A.RR[1] = Bit2+Bit1+Bit0;
   ReadModem=0;

   return;
}

UBYTE SCC_GetB (CPTR addr)
{
   if ((addr & 1) == 0) {
      #ifdef _SCC_Debug
         printf("SCC GB %4lx\n", addr);
      #endif

      switch (addr >> 1) {
         case 0 : // Channel B Control
            if (SCC_Mode == 1) {
               SCC_Mode = 0;
               return SCC_GetBReg();
            } else {
               SCC_Reg = 0;
               return SCC_GetBReg();
            }
            break;

         case 1 : // Channel A Control
            if (SCC_Mode == 1) {
               SCC_Mode = 0;
               return SCC_GetAReg();
            } else {
               SCC_Reg = 0;
               return SCC_GetAReg();
            }
            break;

         case 2 : { // Channel B Data
            #ifdef _SCC_Debug2
               fprintf(stderr,"SCC_GetB - B Data");
            #endif

            if (ReadPrint && (SCC.B.WR[3] & Bit0)) { // Rx Enable
               char PrintByte=PrintData[PrintCount++];

               ReadPrint=1;

               if (PrintCount>PrintBytes-1 || PrintCount>16384) {
                  SCC.B.RR[0]&=~Bit0;
                  ReadPrint=PrintBytes=PrintCount=0;
               }

               return(PrintByte);
            } else
               return 0;

            break;
         }

         case 3 : { // Channel A Data
            #ifdef _SCC_Debug2
               fprintf(stderr,"SCC_GetB - A Data");
            #endif

            if (ReadModem && (SCC.A.WR[3] & Bit0)) { // Rx Enable
               char ModemByte=ModemData[ModemCount++];

               ReadModem=1;

               if (ModemCount>ModemBytes-1 || ModemCount>16384) {
                  SCC.A.RR[0]&=~Bit0;
                  ReadModem=ModemBytes=ModemCount=0;
               }
               return(ModemByte);
            } else
               return 0;
            break;
         }
      }
   } else {
      SCC_Reset();
   }

   return 0;
}

UWORD SCC_GetW (CPTR addr)
{
   #ifdef _SCC_Debug
      printf("Attempted Phase Adjust\n");
   #endif
   return 0;
}

ULONG SCC_GetL (CPTR addr)
{
   #ifdef _SCC_Debug
      printf("Attempted Phase Adjust\n");
   #endif
   return 0;
}

void  SCC_PutB (CPTR addr, UBYTE in)
{
   if ((addr & 1) == 0) {
      #ifdef _SCC_Debug
         printf("SCC PB %4lx %2x\n", addr, in);
      #endif

      switch (addr >> 1) {
         case 0 : // Channel B Control
            if (SCC_Mode == 0) {
               SCC_Mode = 1;
               SCC_Reg = in & 0x0F;
               SCC_PutBReg(in & 0xF0);
            } else {
               SCC_Mode = 0;
               SCC_PutBReg(in);
            }
            break;

         case 1 : // Channel A Control
            if (SCC_Mode == 0) {
               SCC_Mode = 1;
               SCC_Reg = in & 0x0F;
               SCC_PutAReg(in & 0xF0);
            } else {
               SCC_Mode = 0;
               SCC_PutAReg(in);
            }
            break;

         case 2 : { // Channel B Data
            char Temp[255];
            sprintf(Temp,"SCC_PutB - B Data: %d",in);
            fprintf(stderr,Temp);

            if (SCC_B_Out) {
               fwrite(&in,1,1,SCC_B_Out);
               fflush(SCC_B_Out);
            }

            if (PrintPort.PortNumber!=0xff) {
               if (SCC.B.WR[5] & Bit3) { // Tx Enable
                  int Result;

                  Result = Printer_WritePort(PrintPort, in);

                  if (Result)
                     fprintf(stderr, "vMac: Couldn't send to Printer port!\n");
               }
            }

            if (SCC.B.WR[5] & Bit3) { // Tx Enable
               SCC.B.RR[0]|=Bit2; //Tx Buffer Empty  
            } else
               SCC.B.RR[0]&=~Bit2; //Tx Buffer Not Empty

            break;
         }

         case 3 : {    // Channel A Data
            char Temp[255];
            sprintf(Temp,"SCC_PutB - A Data: %d",in);
            fprintf(stderr,Temp);

            if(SCC_A_Out) {
                fwrite(&in,1,1,SCC_A_Out);
                fflush(SCC_A_Out);
            }

            if (ModemPort.PortNumber!=0xff) {
               if (SCC.A.WR[5] & Bit3) { // Tx Enable
                  int Result;

                  Result = Modem_WritePort(ModemPort, in);

                  if (Result!=1)
                     fprintf(stderr,"vMac: Couldn't send to Modem port!\n");
               }
            }

            if (SCC.A.WR[5] & Bit3) { //Tx Enable
               SCC.A.RR[0]|=Bit2; //Tx Buffer Empty
            } else
               SCC.A.RR[0]&=~Bit2; //Tx Buffer Not Empty
            break;
         }
      }
   }
   return;
}

void  SCC_PutW (CPTR addr, UWORD in)
{
   #ifdef _SCC_Debug
      printf("Attemped Phase Adjust\n");
   #endif
}

void  SCC_PutL (CPTR addr, ULONG in)
{
   #ifdef _SCC_Debug
      printf("Attempted Phase Adjust\n");
   #endif
}

UBYTE SCC_GetAReg (void)
{
   #ifdef _SCC_Debug2
      char Temp[255];
      sprintf(Temp,"SCC_GetAReg %d",SCC_Reg);
      fprintf(stderr,Temp);
   #endif

   return SCC.A.RR[SCC_Reg];
}

void  SCC_PutAReg (UBYTE in)
{
   #ifdef _SCC_Debug2
      char Temp[255];
      sprintf(Temp,"SCC_PutAReg %d %d",SCC_Reg, in);
      fprintf(stderr,Temp);
   #endif

   SCC.A.WR[SCC_Reg] = in;
   SCC_PutReg(in,1);
}

UBYTE SCC_GetBReg (void)
{
   #ifdef _SCC_Debug2
      char Temp[255];
      sprintf(Temp,"SCC_GetBReg %d",SCC_Reg);
      fprintf(stderr,Temp);
   #endif

   if (SCC_Reg==2) {
      UBYTE TempData=0;

      if (SCC.B.WR[9] & Bit4) { //Status High
         TempData+=SCC.B.RR[2] && Bit0;
         TempData+=SCC.B.RR[2] && Bit1;
         TempData+=SCC.B.RR[2] && Bit2;
         TempData+=SCC.B.RR[2] && Bit3;
         TempData+=SCC.B.RR[2] && Bit7;

         switch (SCC_Interrupt_Type) {
            case SCC_A_Rx:
               TempData+=Bit4+Bit5;
               break;

            case SCC_A_Rx_Spec:
               TempData+=Bit4+Bit5+Bit6;
               break;

            case SCC_A_Tx_Empty:
               TempData+=Bit4;
               break;

            case SCC_A_Ext:
               TempData+=Bit4+Bit6;
               break;

            case SCC_B_Rx:
               TempData+=Bit5;
               break;

            case SCC_B_Rx_Spec:
               TempData+=Bit5+Bit6;
               break;

            case SCC_B_Tx_Empty:
               TempData+=0;
               break;

            case SCC_B_Ext:
               TempData+=Bit6;
               break;
         }
      } else { //Status Low
         TempData+=SCC.B.RR[2] && Bit0;
         TempData+=SCC.B.RR[2] && Bit4;
         TempData+=SCC.B.RR[2] && Bit5;
         TempData+=SCC.B.RR[2] && Bit6;
         TempData+=SCC.B.RR[2] && Bit7;

         switch (SCC_Interrupt_Type) {
            case SCC_A_Rx:
               TempData+=Bit3+Bit2;
               break;

            case SCC_A_Rx_Spec:
               TempData+=Bit3+Bit2+Bit1;
               break;

            case SCC_A_Tx_Empty:
               TempData+=Bit3;
               break;

            case SCC_A_Ext:
               TempData+=Bit3+Bit1;
               break;

            case SCC_B_Rx:
               TempData+=Bit2;
               break;

            case SCC_B_Rx_Spec:
               TempData+=Bit2+Bit1;
               break;

            case SCC_B_Tx_Empty:
               TempData+=0;
               break;

            case SCC_B_Ext:
               TempData+=Bit1;
               break;
         }
      }

      SCC_Interrupt_Type=0;

      return(TempData);
   }

   return SCC.B.RR[SCC_Reg];
}

void SCC_PutBReg (UBYTE in)
{
   #ifdef _SCC_Debug2
      char Temp[255];
      sprintf(Temp,"SCC_PutBReg %d %d",SCC_Reg, in);
      fprintf(stderr,Temp);
   #endif

   SCC.B.WR[SCC_Reg] = in;
   SCC_PutReg(in,2);

   return;
}

void SCC_Interrupt(int Type)
{
   extern void Exception(int nr, CPTR oldpc);

   #ifdef _SCC_Debug2
      fprintf(stderr,"SCC_Interrupt");
   #endif

   if (SCC.A.WR[9] & Bit3) { //Master Interrupt Enable
      if (Type>SCC_Interrupt_Type)
         SCC_Interrupt_Type=Type;

         Exception(26,1);
   }

   return;
}

void SCC_Int(void)
{
   // This should be called at regular intervals


   // Turn off Sync/Hunt Mode

   if (SCC.A.RR[0] & Bit4) {
      SCC.A.RR[0]&=~Bit4;

      #ifdef _SCC_Debug2
         fprintf(stderr,"SCC_Int: Disable Sync/Hunt on A");
      #endif

      if (SCC.A.RR[15] & Bit4)
         SCC_Interrupt(SCC_A_Ext);
   }

   if (SCC.B.RR[0] & Bit4) {
      SCC.B.RR[0]&=~Bit4;

      #ifdef _SCC_Debug2
         fprintf(stderr,"SCC_Int: Disable Sync/Hunt on B");
      #endif

      if (SCC.B.RR[15] & Bit4)
         SCC_Interrupt(SCC_A_Ext);
   }

   // Check for incoming data
   if (ModemPort.PortNumber!=0xff) {
      if (!(SCC.A.WR[3] & Bit0)) //Rx Disabled
         ReadModem=0;

      if (ModemBytes>0 && (ModemCount>ModemBytes-1)) {
         SCC.A.RR[0]&=~Bit0;
         ReadModem=ModemBytes=ModemCount=0;
      }

      if (ReadModem==1) {
         ReadModem=2;

         SCC.A.RR[0]|=Bit0;

         if (SCC.A.WR[0] & Bit5 && !(SCC.A.WR[0] & (Bit4|Bit3))) //Int on next Rx char
            SCC_Interrupt(SCC_A_Rx);
         else
            if (SCC.A.WR[1] & Bit3 && !(SCC.A.WR[1] & Bit4)) //Int on first Rx char
               SCC_Interrupt(SCC_A_Rx);
            else
               if (SCC.A.WR[1] & Bit4 && !(SCC.A.WR[1] & Bit3)) //Int on all Rx chars
                  SCC_Interrupt(SCC_A_Rx);
      }
   }

   if (PrintPort.PortNumber!=0xff) {
      if (!(SCC.B.WR[3] & Bit0)) //Rx Disabled
         ReadPrint=0;

      if (PrintBytes>0 && (PrintCount>PrintBytes-1)) {
         SCC.B.RR[0]&=~Bit0;
         ReadPrint=PrintBytes=PrintCount=0;
      }

      if (ReadPrint==1) {
         ReadPrint=2;

         SCC.B.RR[0]|=Bit0;

         if (SCC.B.WR[0] & Bit5 && !(SCC.B.WR[0] & (Bit4|Bit3))) //Int on next Rx char
            SCC_Interrupt(SCC_B_Rx);
         else
            if (SCC.B.WR[1] & Bit3 && !(SCC.B.WR[1] & Bit4)) //Int on first Rx char
               SCC_Interrupt(SCC_B_Rx);
            else
               if (SCC.B.WR[1] & Bit4 && !(SCC.B.WR[1] & Bit3)) //Int on all Rx chars
                  SCC_Interrupt(SCC_B_Rx);
      }
   }

   return;
}

void SCC_PutReg(int in, int chan)
{
   // chan 1 is A
   // chan 2 is B

   if (SCC_Reg==2) { //Interrupt Vector (same on RR & WR)
      /*
      Only 1 interrupt vector for the SCC, which
      is stored in channels A and B. B is modified
      when read.
      */
      SCC.A.RR[SCC_Reg]=SCC.B.RR[SCC_Reg]=in;
   }

   if (SCC_Reg==9) { //Only 1 WR9 in the SCC
      SCC.A.WR[SCC_Reg]=SCC.B.WR[SCC_Reg]=in;
   }

   if (SCC_Reg==15) { //Reg 15 is same on RR & WR
      if (chan==1)
         SCC.A.RR[SCC_Reg] = in;
      else
         SCC.B.RR[SCC_Reg] = in;
   }

   if (SCC_Reg==0) {
      if (in>=192) { //Reset Tx Underrun/EOM Latch
         in-=192;
      }

      if (in>=128) { //Reset Tx CRC Generator
         in-=128;
      }

      if (in>=64) { //Reset Rx CRC Checker
         in-=64;
      }

      if (in>=56) { //Reset Highest IUS
         in-=56;
      }

      if (in>=48) { //Error Reset
         if (chan==1) {
            SCC.A.RR[1]&=~Bit6;
            SCC.A.RR[1]&=~Bit5;
            SCC.A.RR[1]&=~Bit4;
         } else {
            SCC.B.RR[1]&=~Bit6;
            SCC.B.RR[1]&=~Bit5;
            SCC.B.RR[1]&=~Bit4;
         }
         in-=48;
      }

      if (in>=40) { //Reset Tx Int Pending
         in-=40;
      }

      if (in>=32) { //Enable Int on next Rx char
         in-=32;
      }

      if (in>=24) { //Send Abort (SDLC)
         if (chan==1) {
            SCC.A.RR[0]|=Bit6;
            SCC.A.RR[0]|=Bit2;
         } else {
            SCC.B.RR[0]|=Bit6;
            SCC.B.RR[0]|=Bit2;
         }
         in-=24;
      }

      if (in>=16) { //Reset Ext/Status Ints
         in-=16;
      }

      if (in>=8) { //Point High
         in-=8;
      }

      if (in>=4) {
         in-=4;
      }

      if (in>=2) {
         in-=2;
      }

      if (in>=1) {
         in-=1;
      }
   }

   if (SCC_Reg==3) {
      if (in<=32) { //Rx 5 Bits/Character
         if (ModemPort.PortNumber!=0xff && chan==1) {
            ModemPort.PortParams.DataBits=5;
            Modem_ReinitPort(ModemPort);
         }
      }

      if (in>=192) { //Rx 8 Bits/Character
         if (ModemPort.PortNumber!=0xff && chan==1) {
            ModemPort.PortParams.DataBits=8;
            Modem_ReinitPort(ModemPort);
         }
         in-=192;
      }  

      if (in>=128) { //Rx 6 Bits/Character
         if (ModemPort.PortNumber!=0xff && chan==1) {
            ModemPort.PortParams.DataBits=6;
            Modem_ReinitPort(ModemPort);
         }
         in-=128;
      }

      if (in>=64) { //Rx 7 Bits/Character
         if (ModemPort.PortNumber!=0xff && chan==1) {
            ModemPort.PortParams.DataBits=7;
            Modem_ReinitPort(ModemPort);
         }
         in-=64;
      }

      if (in>=32) { //Auto Enables
         in-=32;
      }

      if (in>=16) { //Enter Hunt Mode
         if (chan==1) {
            if (!(SCC.A.RR[0] & Bit4)) {
               SCC.A.RR[0]|=Bit4;

               if (SCC.A.RR[15] & Bit4)
                  SCC_Interrupt(SCC_A_Ext);
            }
         } else {
            if (!(SCC.B.RR[0] & Bit4)) {
               SCC.B.RR[0]|=Bit4;

               if (SCC.B.RR[15] & Bit4)
                  SCC_Interrupt(SCC_B_Ext);
            }
         }
         in-=16;
      }

      if (in>=8) { //Rx CRC Enable
         in-=8;
      }

      if (in>=4) { //Addr Search Mode (SDLC)
         in-=4;
      }

      if (in>=2) { //Sync Char Load Inhibit
         in-=2;
      }

      if (in>=1) { //Rx Enable
         in-=1;
      }
   }

   if (SCC_Reg==4) {
      if (in>=128) {
         in-=128;
      }
      if (in>=64) {
         in-=64;
      }
      if (in>=32) {
         in-=32;
      }
      if (in>=16) {
         in-=16;
      }

      if (in>=12) { //2 Stop Bits
         if (ModemPort.PortNumber!=0xff && chan==1) {
            ModemPort.PortParams.StopBits=2;
            Modem_ReinitPort(ModemPort);
         }
         in-=12;
      }

      if (in>=8) { //1 1/2 Stop Bits
         if (ModemPort.PortNumber!=0xff && chan==1) {
            ModemPort.PortParams.StopBits=2;
            Modem_ReinitPort(ModemPort);
         }
         in-=8;
      }

      if (in>=4) { //1 Stop Bit
         if (ModemPort.PortNumber!=0xff && chan==1) {
            ModemPort.PortParams.StopBits=1;
            Modem_ReinitPort(ModemPort);
         }
         in-=4;
      }

      // Parity stuff (bits 0 and 1)
      if (ModemPort.PortNumber!=0xff && chan==1) {
         if (in & Bit0) {
              if (in & Bit1) //SCC only supports even and odd parities
                 ModemPort.PortParams.Parity=2;
              else
                 ModemPort.PortParams.Parity=1;
         } else
            ModemPort.PortParams.Parity=0;

         Modem_ReinitPort(ModemPort);
      }
   }

   if (SCC_Reg==9) {
      if (in>=192) {
         SCC_Reset();
         in-=192;
      }
      if (in>=128) {
         SCC_ResetA();
         in-=128;
      }
      if (in>=64) {
         SCC_ResetB();
         in-=64;
      }
      if (in>=32) {
         in-=32;
      }
      if (in>=16) {
         in-=16;
      }
      if (in>=8) {
         in-=8;
      }
      if (in>=4) {
         in-=4;
      }
      if (in>=2) {
         in-=2;
      }
      if (in>=1) {
         in-=1;
      }
   }

   //Baud Rate
   if ((SCC_Reg==12 || SCC_Reg==13) && ((chan==1 && ModemPort.PortNumber!=0xff) || (chan==2 && PrintPort.PortNumber!=0xff))) {
      switch (SCC.A.WR[12]+(SCC.A.WR[13]<<8)) {
         case 380:
            ModemPort.PortParams.BaudRate=300;
            break;

         case 94:
            ModemPort.PortParams.BaudRate=1200;
            break;

         case 46:
            ModemPort.PortParams.BaudRate=2400;
            break;

         case 22:
            ModemPort.PortParams.BaudRate=4800;
            break;

         case 10:
            ModemPort.PortParams.BaudRate=9600;
            break;

         default:
            ModemPort.PortParams.BaudRate=9600;
            break;
      }


      if (chan==1)
         Modem_ReinitPort(ModemPort);
   }

   return;
}

// VIA Interface Functions

UBYTE VIA_GORA7 (void)
{
  return 1; // No Wait/Requests
}

void  VIA_PORA7 (UBYTE Data)
{
  #ifdef _VIA_Interface_Debug
  printf("VIA ORA7 attempts to be an output\n");
  #endif
}
